import cooler
import numpy as np
import pandas as pd
from datetime import datetime
import multiprocessing
from functools import partial
import warnings
from scipy.stats import binom
from scipy.sparse import csr_matrix
import scipy
import sys
import pickle
import gc
##open files
import h5py
import scipy.sparse as sparse
warnings.filterwarnings('ignore')
from additional_finctions import chr_pair_h5loader
import logging
logging.basicConfig(level = logging.INFO)


def calculate_sweet_sum_on_diag_sparse(diag):
    coords = []
    data = []
    if diag%100==0:
        logging.getLogger(__name__).info(str(diag) + str(datetime.now()))

    for i in range(diag, diff_array.shape[0], step):
        left_sweet_square = diff_array_csr[i-diag-sweet_size:i-diag, i-sweet_size:i].toarray()
        right_sweet_square = diff_array_csr[i-diag:i-diag+sweet_size, i:i+sweet_size].toarray() 
        
        # ##right_sweet
        left_sweet_sum = np.sum(np.triu(np.rot90(left_sweet_square, 1)))
        right_sweet_sum = np.sum(np.triu(np.rot90(right_sweet_square, 3)))
        coords.append((i-diag, i))
        data.append(left_sweet_sum+right_sweet_sum)
    return diag, coords, data

global diff_array_csr, use_only_p_c, sweet_size, step, out_dir, cool_ref, sample_path

sample = sys.argv[1]
sample_path = sys.argv[2]
chrom_search = sys.argv[4]
bin_size = int(sys.argv[3])
sweet_size = int(sys.argv[5])
big_depth_sample = int(sys.argv[6])
big_depth_ref = int(sys.argv[7])

n_cpus = 10
step = 1
res_size_data = pd.read_csv("./res_size.txt", sep="\t")

logging.getLogger(__name__).info(sample + "start calculate sweet for chrom "+chrom_search+" with bin size "+str(bin_size))
use_only_p_c = True
out_dir = "./temp/"
chroms = list(range(1,23))
chroms[:] = ["chr" + str(chr_number) for chr_number in chroms]
ref_path = "./sup_pat.h5"
cool_ref = cooler.Cooler("./samples/sup_pat.mcool::resolutions/"+str(bin_size))
# cool_sample = cooler.Cooler(sample_path)
##get depth for all chromosomes

min_diag = res_size_data[res_size_data["resolution"]==bin_size].iloc[0,1]//bin_size
max_diag = res_size_data[res_size_data["resolution"]==bin_size].iloc[0,2]//bin_size
## open hic data for patient and control in sparse format
sample_data_raw = chr_pair_h5loader(sample_path, chrom_search+"-"+chrom_search, bin_size).tocoo()
# ref_data_raw = chr_pair_h5loader(ref_path, chrom_search+"-"+chrom_search, bin_size).tocoo()
ref_data_raw = scipy.sparse.triu(cool_ref.matrix(balance=False, sparse=True).fetch(chrom_search))
logging.getLogger(__name__).info("find same coord")
if use_only_p_c == True:
    max_coord_const = np.int64(10**7)
    _,_, same_coords_inds = np.intersect1d(np.int64(sample_data_raw.row)*max_coord_const+np.int64(sample_data_raw.col),
    np.int64(ref_data_raw.row)*max_coord_const+np.int64(ref_data_raw.col), assume_unique=True, return_indices=True)
    if len(same_coords_inds) != len(sample_data_raw.data):
        ref_data_raw = (ref_data_raw+sample_data_raw).tocoo()
        _,_, same_coords_inds = np.intersect1d(np.int64(sample_data_raw.row)*max_coord_const+np.int64(sample_data_raw.col),
        np.int64(ref_data_raw.row)*max_coord_const+np.int64(ref_data_raw.col), assume_unique=True, return_indices=True)
    ref_data_raw.data = ref_data_raw.data[same_coords_inds]
    ref_data_raw.row = ref_data_raw.row[same_coords_inds]
    ref_data_raw.col = ref_data_raw.col[same_coords_inds]
assert ref_data_raw.data.shape ==sample_data_raw.data.shape
logging.getLogger(__name__).info("calc binom")
sample_data_raw.data = binom.pmf(sample_data_raw.data, big_depth_sample, ref_data_raw.data/big_depth_ref)
ref_data_raw.data = binom.pmf(ref_data_raw.data, big_depth_ref, ref_data_raw.data/big_depth_ref)
diff_array = sample_data_raw
logging.getLogger(__name__).info("log")
diff_array.data = np.log(sample_data_raw.data/ref_data_raw.data)
logging.getLogger(__name__).info("create diff_array_csr")
diff_array_csr = diff_array.tocsr()
logging.getLogger(__name__).info("create sweet")
if max_diag > diff_array_csr.get_shape()[0]-sweet_size:
    max_diag =  diff_array_csr.get_shape()[0]-sweet_size
pool = multiprocessing.Pool(n_cpus)
logging.getLogger(__name__).info("bin_size " + str(bin_size) + " min_diag " + str(min_diag) + " max_diag " + str(max_diag))
result = pool.map(partial(calculate_sweet_sum_on_diag_sparse), range(min_diag,max_diag))
row, col, data =[], [], []
for diag_coord_data in result:
    data+=diag_coord_data[2]
    row_cols = [[i for i, j in diag_coord_data[1]],
                [j for i, j in diag_coord_data[1]]]
    row+=row_cols[0]
    col+=row_cols[1]
with open(out_dir + sample + "_sweet_"+str(sweet_size)+"_"+str(bin_size//1000)+"Kb_"+chrom_search+"_wo_edges.pickle", 'wb') as f:
    pickle.dump( np.array([np.array(row), np.array(col), np.array(data)]), f)

del diff_array_csr
del result
del sample_data_raw
del ref_data_raw
gc.collect()
del cool_ref
gc.collect()

logging.getLogger(__name__).info("end " + str(datetime.now()))
